!---------------------------------------------------------------------------------------
! PREP_FLUXES: mod_settings
!---------------------------------------------------------------------------------------
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!
!  Copyright 2017, Rona Thompson
!---------------------------------------------------------------------------------------
!
!> mod_settings
!!
!! Purpose:    Constructs the settings data structure.
!!
!---------------------------------------------------------------------------------------

module mod_settings

  use mod_var

  implicit none
  private

  public :: settings_t, read_file_settings, init_settings

  type :: settings_t
  
    character(len=max_path_len)  :: file_in, file_out, file_scalar
    character(len=max_name_len)  :: lonname_in, latname_in, timename_in, varname_in, varunit
    character(len=max_name_len)  :: lonname_scalar, latname_scalar, timename_scalar, varname_scalar
    character(len=max_name_len)  :: lonname_out, latname_out, timename_out, varname_out
    character(len=max_name_len)  :: timestamp
    logical                      :: midpoints  
    real                         :: coeff
    integer                      :: nx_in, ny_in, nx_out, ny_out, ntime, indextime, ntime_out, year
    real                         :: dx_out, dy_out, llx_out, lly_out
    integer(kind=8)              :: timeref
    

  end type settings_t

  contains

    ! --------------------------------------------------
    ! check if input is numeric
    ! --------------------------------------------------

    logical function is_numeric(string)

      character(len=200), intent(in) :: string
      real :: var
      integer :: ierr

      read(string,*,iostat=ierr) var
      if(ierr.eq.0) then
        is_numeric=.true.
      else
        is_numeric=.false.
      endif

    end function is_numeric

    ! --------------------------------------------------
    ! check if input is logical
    ! --------------------------------------------------

    logical function is_logical(string)

      character(len=200), intent(in) :: string
      logical :: var
      integer :: ierr

      read(string,*,iostat=ierr) var
      if(ierr.eq.0) then
        is_logical=.true.
      else
        is_logical=.false.
      endif

    end function is_logical

    ! --------------------------------------------------
    ! split string
    ! --------------------------------------------------

    subroutine split(str,delims,before,sep)

    ! Routine finds the first instance of a character from 'delims' in the
    ! the string 'str'. The characters before the found delimiter are
    ! output in 'before'. The characters after the found delimiter are
    ! output in 'str'. The optional output character 'sep' contains the 
    ! found delimiter. A delimiter in 'str' is treated like an ordinary 
    ! character if it is preceded by a backslash (\). If the backslash 
    ! character is desired in 'str', then precede it with another backslash.

    character(len=*) :: str, delims, before
    character,optional :: sep
    logical :: pres
    character :: ch,cha
    integer :: i, k, ibsl, ipos, iposa, lenstr

    pres=present(sep)
    str=adjustl(str)
    lenstr=len_trim(str)
    if(lenstr == 0) return        ! string str is empty
    k=0
    ibsl=0                        ! backslash initially inactive
    before=' '
    do i=1,lenstr
      ch=str(i:i)
      if(ibsl == 1) then          ! backslash active
        k=k+1
        before(k:k)=ch
        ibsl=0
        cycle
      end if
      if(ch == '\') then          ! backslash with backslash inactive
        k=k+1
        before(k:k)=ch
        ibsl=1
        cycle
      end if
      ipos=index(delims,ch)
      if(ipos == 0) then          ! character is not a delimiter
        k=k+1
        before(k:k)=ch
        cycle
      end if
      if(ch /= ' ') then          ! character is a delimiter that is not a space
        str=str(i+1:)
        if(pres) sep=ch
        exit
      end if
      cha=str(i+1:i+1)            ! character is a space delimiter
      iposa=index(delims,cha)
      if(iposa > 0) then          ! next character is a delimiter
        str=str(i+2:)
        if(pres) sep=cha
        exit
      else
        str=str(i+1:)
        if(pres) sep=ch
        exit
      end if
    end do
    if(i >= lenstr) str=''
    str=trim(str)
    before=trim(before)
    return

    end subroutine split

    ! --------------------------------------------------
    ! routine to identify and read in content
    ! --------------------------------------------------

    subroutine read_content (line, identifier, cc, cn, cl, match)

      character (len=200), intent (in) :: line, identifier
      character (len=200), intent (out) :: cc ! character content
      real(kind=8),        intent (out) :: cn ! numeric content
      logical,             intent (out) :: cl ! logical content
      logical,             intent (out) :: match

      integer :: n

      n = len_trim (identifier)

      ! default: line does not match, cc="", cn=-9900.0
      cc    = ""
      cn    = -9999.9
      match = .false.
      if ( len_trim (line) >= n ) then
        if ( line(:n) == identifier(:n) ) then
          cc = adjustl (line(n+1:))
          if ( is_numeric(cc) ) then
            read (line(n+1:),*) cn
          else
            if( is_logical(cc) ) read (line(n+1:),*) cl
          endif
          match = .true.
        end if
      end if

    end subroutine read_content

    ! --------------------------------------------------
    ! routine to parse a string
    ! --------------------------------------------------

    subroutine parse_string (str, delims, args, nargs)

    ! Parses the string 'str' into arguments args(1), ..., args(nargs) based on
    ! the delimiters contained in the string 'delims'. Preceding a delimiter in
    ! 'str' by a backslash (\) makes this particular instance not a delimiter.
    ! The integer output variable nargs contains the number of arguments found.

    character(len=*) :: str, delims
    character(len=len_trim(str)) :: strsav
    character(len=*), dimension(:) :: args
    integer :: i, na, lenstr, nargs

    strsav=str
    na=size(args)

    do i=1,na
      args(i)=' '
    end do
    nargs=0
    lenstr=len_trim(str)
    if(lenstr==0) return

    do
      if(len_trim(str) == 0) exit
      nargs=nargs+1
      call split(str,delims,args(nargs))
    end do
    str=strsav

    end subroutine parse_string

    ! --------------------------------------------------
    ! read file settings
    ! --------------------------------------------------

    subroutine read_file_settings(filename, settings)

      character(len=max_path_len), intent(in)     :: filename
      type(settings_t),            intent(in out) :: settings

      integer            :: ierr, i, n
      character(len=200) :: line, identifier, cc
      real(kind=8)       :: cn
      logical            :: match, cl

     ! initialization
     settings%file_scalar = ''
     settings%timestamp = ''
  
     ! open file
      open (100, file = trim (filename), status = 'old', iostat=ierr)
      if(ierr.gt.0) then
        write (*,*) 'ERROR: cannot find SETTINGS_files'
        stop
      endif

      ! loop over lines
      read_loop: do
        read (100, fmt='(A)', iostat=ierr) line

        ! first check if line exists or eof was reached
        if ( ierr.gt.0 ) then
          write (*,*) 'ERROR: check ', filename
          exit read_loop
        else if ( ierr.lt.0 ) then
          exit read_loop
        else

          ! omit comments ("#") and empty lines
          if ( index (trim (line), "#") .eq. 1 ) cycle read_loop
          if ( len   (trim (line))      .eq. 0 ) cycle read_loop

          ! read input file settings
          identifier = "file_in:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%file_in = cc
          identifier = "varname_in:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%varname_in = cc
          identifier = "coeff:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%coeff = cn
          identifier = "lonname_in:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%lonname_in = cc
          identifier = "latname_in:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%latname_in = cc
          identifier = "timename_in:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%timename_in = cc
          identifier = "midpoints:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%midpoints = cl
          identifier = "nx_in:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%nx_in = cn
          identifier = "ny_in:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%ny_in = cn
          identifier = "ntime:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%ntime = cn
          identifier = "indextime:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%indextime = cn
          identifier = "timeref:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%timeref = cn
          identifier = "timestamp:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%timestamp = cc

          ! read output file settings
          identifier = "file_out:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%file_out = cc
          identifier = "varname_out:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%varname_out = cc
          identifier = "varunit:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%varunit = cc
          identifier = "lonname_out:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%lonname_out = cc
          identifier = "latname_out:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%latname_out = cc
          identifier = "timename_out:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%timename_out = cc
          identifier = "year:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%year = cn
          identifier = "nx_out:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%nx_out = cn
          identifier = "ny_out:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%ny_out = cn
          identifier = "dx_out:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%dx_out = cn
          identifier = "dy_out:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%dy_out = cn
          identifier = "ntime_out:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%ntime_out = cn
          identifier = "llx_out:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%llx_out = cn
          identifier = "lly_out:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%lly_out = cn

          ! if file_scalar is specified
          identifier = "file_scalar:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%file_scalar = cc
          identifier = "varname_scalar:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%varname_scalar = cc
          identifier = "lonname_scalar:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%lonname_scalar = cc
          identifier = "latname_scalar:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%latname_scalar = cc
          identifier = "timename_scalar:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%timename_scalar = cc

        endif

      end do read_loop

      ! for testing
      print*, trim(settings%file_in)
      print*, trim(settings%varname_in)
      print*, trim(settings%lonname_in)
      print*, trim(settings%latname_in)
      print*, trim(settings%timename_in)
      print*, settings%midpoints
      print*, settings%nx_in
      print*, settings%ny_in
      print*, settings%ntime
      print*, settings%timeref
      print*, trim(settings%file_out)
      print*, trim(settings%varname_out)
      print*, trim(settings%lonname_out)
      print*, trim(settings%latname_out)
      print*, trim(settings%timename_out)
      print*, settings%year
      print*, settings%nx_out
      print*, settings%ny_out
      print*, settings%dx_out  
      print*, settings%dy_out  
      print*, settings%ntime_out
      print*, settings%llx_out  
      print*, settings%lly_out  
      if(len_trim(settings%file_scalar).ne.0) print*, settings%file_scalar

    end subroutine read_file_settings

    ! --------------------------------------------------
    ! initialize settings variables
    ! --------------------------------------------------

    subroutine init_settings(settings)

      use mod_var

      implicit none

      type(settings_t),            intent(in out) :: settings

      nx_in = settings%nx_in
      ny_in = settings%ny_in
      ntime = settings%ntime
      nx_out = settings%nx_out
      ny_out = settings%ny_out
      dx_out = settings%dx_out
      dy_out = settings%dy_out
      llx_out = settings%llx_out
      lly_out = settings%lly_out
      ntime_out = settings%ntime_out
      
      ! for testing
      print*, 'nx_in, ny_in:', nx_in, ny_in
      print*, 'ntime:', ntime
      print*, 'nx_out, ny_out:', nx_out, ny_out
      print*, 'dx_out, dy_out:', dx_out, dy_out
      print*, 'llx_out, lly_out:', llx_out, lly_out

    end subroutine init_settings
 
    ! --------------------------------------------------


end module mod_settings

